// Load in dependencies
var fs = require('fs');
var path = require('path');
var _ = require('lodash');
var async = require('async');
var gruntRetro = require('grunt-retro');
var request = require('request');
var tar = require('tar');

// Define our tasks
module.exports = function (grunt) {
  // Load in grunt-retro
  grunt = gruntRetro(grunt);

  // Define a task 
  grunt.registerMultiTask('getRepo', 'Download collections of files from the internet via grunt.', function () {

    // Collect the filepaths we need
    var src = this.file.src;
    var dest = this.file.config.dest;
    var done = this.async();
    // set extension 
    var extension = !this.file.config.extension ? 'tgz' : this.file.config.extension;

    //collect configuration
    var config = {
      extension: extension,
      repositoryUrl: (this.file.config.repository.url),
      proxy: (typeof this.file.config.repository.proxy === "undefined" ? "" : this.file.config.repository.proxy),
      strictSSL: (typeof this.file.config.repository['strict-ssl'] === "undefined" ? false: this.file.config.repository['strict-ssl']),
      auth: (typeof this.file.config.repository.auth === "undefined" ? false : {username: this.file.config.repository.auth.username, password: this.file.config.repository.auth.password})
    };

    // map files to download
    var files = []
    for (i = 0; i < src.length; i++) {
      var file = (!src[i].url ? config.repositoryUrl : src[i].url) + '/' + src[i].package + '/-/' + src[i].package + '-' + src[i].version + '.' + config.extension;
      files.push(file);
    }

    var srcFiles = files;

    // Iterate over the array and expand the braces
    var minimatch = grunt.file.glob.minimatch,
      braceExpand = minimatch.braceExpand;
    srcFiles = srcFiles.reduce(function expandSrcFiles(retArr, srcFile) {
      var srcFileArr = typeof srcFile === 'string' ? braceExpand(srcFile) : [srcFile];
      retArr = retArr.concat(srcFileArr);
      return retArr;
    }, []);

    // Determine the destinations
    var router = this.data.router || function defaultRouter(filepath) {
      if (typeof filepath !== 'string') {
        filepath = filepath.url || filepath.uri;
      }
      return path.basename(filepath);
    };
    var fileInfos = srcFiles.map(function getDest(srcFile, i) {
      // Route the file, append it to dest, and return
      var filepath = router(srcFile),
        retStr = path.join(dest, filepath);
      return {
        src: srcFile,
        dest: retStr,
        proxy: config.proxy,
        strictSSL: config.strictSSL,
        auth: config.auth
      };
    });

    // Asynchronously fetch the files in parallel
    async.map(fileInfos, grunt.helper.bind(grunt, 'getRepo'), function handleGetRepoResult(err) {
      // If there is an error, fail
      if (err) {
        grunt.fail.warn(err);
        return done();
      }
      // Otherwise, print a success message.
      var destArr = fileInfos.map(function (item) {
        return item.dest;
      })

      // setup patches and message
      destArr.forEach(function (item, index) {
        var dir = dest + '/' + src[index].package;
        var mess = src[index].package + '-' + src[index].version;
        var copy = src[index].copy;
        if (!fs.existsSync(dir)) {
          fs.mkdirSync(dir);
        }
        //extract files
        tar.x(
          {
            file: item,
            cwd: dir,
            sync: true,
            strip: true,
            unlink: true
          }
        )
        // delete packed  file
        fs.unlinkSync(item);

        //copy files
        if (copy) {
          copy.forEach(function (item, index) {
            let filesToCopy = [];
            let src = checkFile(dir + '/' + item.src);
            if (src !== 'file' && item.dest.substr(-1) !== '/') {
              item.dest = item.dest + '/';
            }
            let dest = item.dest.substr(0, item.dest.lastIndexOf("/"));
            checkDir(dest);
            let fromsrc = item.src.substr(0, item.src.lastIndexOf("/"));
            let dirToDownload = '';
            if (fromsrc) {
              dirToDownload = '/' + fromsrc;
            }
            switch (src) {
              //copy one file
              case 'file':
                filesToCopy.push({ src: dir + '/' + item.src, dest: item.dest })
                break;
              //copy all files from a directory
              case 'dir':
              case 'all':
                if (checkFile(item.dest) === 'dir') {
                  filesToCopy = copyFromDir(dir + dirToDownload, 'all', item.dest);
                }
                else message('Error to copy files to ' + item.dest + ' ! Destionation must be a valid directory.');
                break;
              //copy files based on extension
              default:
                if (checkFile(item.dest) === 'dir') {
                  filesToCopy = copyFromDir(dir + dirToDownload, src, item.dest);
                }
                else message('Error to copy files to ' + item.dest + ' ! Destionation must be a valid directory.');
                break;
            }
            filesToCopy.forEach(function (filetocopy) {
              fs.copyFileSync(filetocopy.src, filetocopy.dest, (err) => {
                if (err) throw err;
              });
            })
          })
        }
        // write message
        grunt.log.writeln('+ ' + mess);
      })
      // Callback
      done();
    });
  });

  // Register our getRepo helper
  grunt.registerHelper('getRepo', function (info, cb) {
    // Default to a binary request
    var options = {
      url: info.src,
      proxy: info.proxy,
      strictSSL: info.strictSSL,
      auth: info.auth
    }
    var dest = info.dest;

    // Request the url
    var req = request(options);

    // On error, callback
    req.on('error', cb);

    // On response, callback for writing out the stream
    req.on('response', function handleResponse(res) {
      // Assert the statusCode was good
      var statusCode = res.statusCode;
      if (statusCode < 200 || statusCode >= 300) {
        return cb(new Error('Fetching ' + JSON.stringify(options) + ' failed with HTTP status code ' + statusCode));
      }
      // Otherwise, write out the content
      var destdir = path.dirname(dest);
      grunt.file.mkdir(destdir);
      var writeStream = fs.createWriteStream(dest);

      var dataStream = req;
      dataStream.pipe(writeStream);

      // When the stream errors or completes, exit
      writeStream.on('error', cb);
      writeStream.on('close', cb);
    });
  });
};

//check the directory and create if not exists
function checkDir(dir) {
  let dirArr = dir.split('/');
  let dirToCheck = ""
  dirArr.forEach(function (item, index) {
      if (index === 0) {
          dirToCheck = item;
      }
      else {
          dirToCheck = dirToCheck + '/' + item;
      }

      if (!fs.existsSync(dirToCheck)) {
          fs.mkdirSync(dirToCheck);
      }
  })

}

function copyFromDir(src, extension, dest) {
  const directoryPath = path.join(__dirname, '../../../' + src);
  let ret = [];
  let files = fs.readdirSync(directoryPath);
  files.forEach(function (file) {
    let ext = path.extname(file);
    let isFile = checkFile(directoryPath + '\\' + file);
    if (isFile === 'file' && (extension === 'all' || extension === ext)) {
      if (dest.substr(-1) !== '/') {
        dest = dest + '/';
      }
      ret.push({ src: src + '/' + file, dest: dest + file });
    }
  });
  return ret;
}

function checkFile(file) {
  var ret = null;
  var ext = path.extname(file);
  var filename = path.basename(file);
  if (filename === '*.*') {
    ret = 'all';
  }
  else if (filename.substr(0, 2) === "*.") {
    ret = ext;
  }
  if (!ret && fs.existsSync(file)) {
    let stats = fs.statSync(file);
    if (stats) {
      if (stats.isFile()) {
        ret = 'file';
      }
      if (stats.isDirectory()) {
        ret = 'dir';
      }
    }
  }
  return ret;
}

// message printout
function message(message) {
  console.log(message);
}
